import abc
import time
import types

import six
import theano
import theano.sparse
import theano.tensor as T

from neupy import layers
from neupy.utils import AttributeKeyDict, asfloat, format_data, as_tuple
from neupy.layers.utils import preformat_layer_shape, iter_parameters
from neupy.layers.connections import LayerConnection, is_sequential
from neupy.layers.connections.base import create_input_variables
from neupy.exceptions import InvalidConnection
from neupy.core.properties import ChoiceProperty
from neupy.algorithms.base import BaseNetwork
from .gd import errors


__all__ = ('ConstructibleNetwork',)


def does_layer_accept_1d_feature(layer):
    """
    Check if 1D feature values are valid for the layer.

    Parameters
    ----------
    layer : object

    Returns
    -------
    bool
    """
    return (layer.output_shape == (1,))


def generate_layers(layers_sizes):
    """
    Create from list of layer sizes basic linear network.

    Parameters
    ----------
    layers_sizes : list or tuple
        Ordered list of network connection structure.

    Returns
    -------
    LayerConnection
        Constructed connection.
    """
    layers_sizes = list(layers_sizes)
    n_layers = len(layers_sizes)

    if n_layers <= 1:
        raise ValueError("Cannot generate network that "
                         "has less than two layers")

    input_layer_size = layers_sizes.pop(0)
    connection = layers.Input(input_layer_size)

    for output_size in layers_sizes:
        next_layer = layers.Sigmoid(output_size)
        connection = LayerConnection(connection, next_layer)

    return connection


def clean_layers(connection):
    """
    Clean layers connections and format transform them into one format.
    Also this function validate layers connections.

    Parameters
    ----------
    connection : list, tuple or object
        Layers connetion in different formats.

    Returns
    -------
    object
        Cleaned layers connection.
    """
    if all(isinstance(element, int) for element in connection):
        connection = generate_layers(connection)

    if isinstance(connection, (list, tuple)):
        connection = layers.join(*connection)

    return connection


def create_output_variable(error_function, name):
    """
    Create output variable based on error function.

    Parameters
    ----------
    error_function : function
    name : str

    Returns
    -------
    Theano variable
    """
    # TODO: Solution is not user friendly. I need to find
    # better solution later.
    if hasattr(error_function, 'expected_dtype'):
        network_output_dtype = error_function.expected_dtype
    else:
        network_output_dtype = T.matrix

    return network_output_dtype(name)


class ErrorFunctionProperty(ChoiceProperty):
    """
    Property that helps select error function from
    available or define a new one.

    Parameters
    ----------
    {ChoiceProperty.Parameters}
    """
    def __set__(self, instance, value):
        if isinstance(value, types.FunctionType):
            return super(ChoiceProperty, self).__set__(instance, value)
        return super(ErrorFunctionProperty, self).__set__(instance, value)

    def __get__(self, instance, value):
        founded_value = super(ChoiceProperty, self).__get__(instance, value)
        if isinstance(founded_value, types.FunctionType):
            return founded_value
        return super(ErrorFunctionProperty, self).__get__(instance,
                                                          founded_value)


class BaseAlgorithm(six.with_metaclass(abc.ABCMeta)):
    """
    Base class for algorithms implemeted in Theano.

    Attributes
    ----------
    variables : dict
        Theano variables.

    methods : dict
        Compiled Theano functions.
    """
    def __init__(self, *args, **kwargs):
        super(BaseAlgorithm, self).__init__(*args, **kwargs)

        self.logs.message("THEANO", "Initializing Theano variables and "
                                    "functions.")
        start_init_time = time.time()

        self.variables = AttributeKeyDict()
        self.methods = AttributeKeyDict()

        self.init_input_output_variables()
        self.init_variables()
        self.init_methods()

        finish_init_time = time.time()
        self.logs.message("THEANO", "Initialization finished successfully. "
                          "It took {:.2f} seconds"
                          "".format(finish_init_time - start_init_time))

    @abc.abstractmethod
    def init_input_output_variables(self):
        """
        Initialize input and output Theano variables.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def init_variables(self):
        """
        Initialize Theano variables.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def init_methods(self):
        """
        Initialize Theano functions.
        """
        raise NotImplementedError


class ConstructibleNetwork(BaseAlgorithm, BaseNetwork):
    """
    Class contains functionality that helps work with network that have
    constructible layers architecture.

    Parameters
    ----------
    connection : list, tuple or LayerConnection instance
        Network's architecture. There are a few ways
        to define it.

        - List of layers.
          For instance, ``[Input(2), Tanh(4), Relu(1)]``.

        - Construct layer connections.
          For instance, ``Input(2) > Tanh(4) > Relu(1)``.

        - Tuple of integers. Each integer defines Sigmoid layer
          and it's input size.  For instance,  value ``(2, 4, 1)``
          means that network has 3 layers with 2 input units,
          4 hidden units and 1 output unit.

    error : str or function
        Error/loss function. Defaults to ``mse``.

        - ``mae`` - Mean Absolute Error.

        - ``mse`` - Mean Squared Error.

        - ``rmse`` - Root Mean Squared Error.

        - ``msle`` - Mean Squared Logarithmic Error.

        - ``rmsle`` - Root Mean Squared Logarithmic Error.

        - ``categorical_crossentropy`` - Categorical cross entropy.

        - ``binary_crossentropy`` - Binary cross entropy.

        - ``binary_hinge`` - Binary hinge entropy.

        - ``categorical_hinge`` - Categorical hinge entropy.

        - Custom function which accepts two mandatory arguments.
          The first one is expected value and the second one is
          predicted value. Example:

        .. code-block:: python

            def custom_func(expected, predicted):
                return expected - predicted

    {BaseNetwork.Parameters}

    Attributes
    ----------
    {BaseNetwork.Attributes}

    Methods
    -------
    {BaseSkeleton.predict}

    train(input_train, target_train, input_test=None, target_test=None,\
    epochs=100, epsilon=None)
        Train network. You can control network's training procedure
        with ``epochs`` and ``epsilon`` parameters.
        The ``input_test`` and ``target_test`` should be presented
        both in case of you need to validate network's training
        after each iteration.

    {BaseSkeleton.fit}
    """
    error = ErrorFunctionProperty(default='mse', choices={
        'mae': errors.mae,
        'mse': errors.mse,
        'rmse': errors.rmse,
        'msle': errors.msle,
        'rmsle': errors.rmsle,

        'binary_crossentropy': errors.binary_crossentropy,
        'categorical_crossentropy': errors.categorical_crossentropy,

        'binary_hinge': errors.binary_hinge,
        'categorical_hinge': errors.categorical_hinge,
    })

    def __init__(self, connection, *args, **kwargs):
        self.connection = clean_layers(connection)

        self.layers = list(self.connection)
        graph = self.connection.graph

        if len(self.connection.output_layers) != 1:
            n_outputs = len(graph.output_layers)
            raise InvalidConnection("Connection should have one output "
                                    "layer, got {}".format(n_outputs))

        self.output_layer = graph.output_layers[0]

        super(ConstructibleNetwork, self).__init__(*args, **kwargs)

    def init_input_output_variables(self):
        self.variables.update(
            network_inputs=create_input_variables(
                self.connection.input_layers),
            network_output=create_output_variable(
                self.error,
                name='algo:network/var:network-output',
            ),
        )

    def init_variables(self):
        network_inputs = self.variables.network_inputs
        network_output = self.variables.network_output

        train_prediction = self.connection.output(*network_inputs)
        with self.connection.disable_training_state():
            prediction = self.connection.output(*network_inputs)

        self.variables.update(
            step=theano.shared(
                name='algo:network/scalar:step',
                value=asfloat(self.step)
            ),
            epoch=theano.shared(
                name='algo:network/scalar:epoch',
                value=asfloat(self.last_epoch)
            ),

            prediction_func=prediction,
            train_prediction_func=train_prediction,

            error_func=self.error(network_output, train_prediction),
            validation_error_func=self.error(network_output, prediction),
        )

    def init_methods(self):
        network_inputs = self.variables.network_inputs
        network_output = self.variables.network_output

        self.methods.update(
            predict=theano.function(
                inputs=network_inputs,
                outputs=self.variables.prediction_func,
                name='algo:network/func:predict'
            ),
            train_epoch=theano.function(
                inputs=network_inputs + [network_output],
                outputs=self.variables.error_func,
                updates=self.init_train_updates(),
                name='algo:network/func:train-epoch'
            ),
            prediction_error=theano.function(
                inputs=network_inputs + [network_output],
                outputs=self.variables.validation_error_func,
                name='algo:network/func:prediction-error'
            )
        )

    def init_train_updates(self):
        """
        Initialize updates that would be applied after
        each training epoch.
        """
        updates = []
        for layer, _, parameter in iter_parameters(self.layers):
            updates.extend(self.init_param_updates(layer, parameter))

        for layer in self.layers:
            updates.extend(layer.updates)

        return updates

    def init_param_updates(self, layer, parameter):
        """
        Initialize parameter updates.

        Parameters
        ----------
        layer : object
            Any layer that inherit from BaseLayer class.
        parameter : object
            Usualy it is a weight or bias.

        Returns
        -------
        list
            List of updates related to the specified parameter.
        """
        return []

    def format_input_data(self, input_data):
        """
        Input data format is depend on the input layer
        structure.

        Parameters
        ----------
        input_data : array-like or None

        Returns
        -------
        array-like or None
            Function returns formatted array.
        """
        input_layers = self.connection.input_layers

        if not isinstance(input_data, (tuple, list)):
            input_layer = input_layers[0]
            is_feature1d = does_layer_accept_1d_feature(input_layer)
            return format_data(input_data, is_feature1d)

        formated_data = []
        for input_to_layer, input_layer in zip(input_data, input_layers):
            is_feature1d = does_layer_accept_1d_feature(input_layer)
            data = format_data(input_to_layer, is_feature1d)
            formated_data.append(data)

        return tuple(formated_data)

    def format_target_data(self, target_data):
        """
        Target data format is depend on the output layer
        structure.

        Parameters
        ----------
        target_data : array-like or None

        Returns
        -------
        array-like or None
            Function returns formatted array.
        """
        return format_data(target_data)

    def prediction_error(self, input_data, target_data):
        """
        Calculate prediction accuracy for input data.

        Parameters
        ----------
        input_data : array-like
        target_data : array-like

        Returns
        -------
        float
            Prediction error.
        """
        return self.methods.prediction_error(*as_tuple(
            self.format_input_data(input_data),
            self.format_target_data(target_data)
        ))

    def predict(self, input_data):
        """
        Return prediction results for the input data.

        Parameters
        ----------
        input_data : array-like

        Returns
        -------
        array-like
        """
        input_data = self.format_input_data(input_data)
        return self.methods.predict(*as_tuple(input_data))

    def on_epoch_start_update(self, epoch):
        """
        Function will be triggered before the training epoch procedure.

        Parameters
        ----------
        epoch : int
            Current epoch number.
        """
        super(ConstructibleNetwork, self).on_epoch_start_update(epoch)
        self.variables.epoch.set_value(epoch)

    def train(self, input_train, target_train, input_test=None,
              target_test=None, *args, **kwargs):
        """
        Train neural network.
        """
        is_test_data_partialy_missed = (
            (input_test is None and target_test is not None) or
            (input_test is not None and target_test is None)
        )

        if is_test_data_partialy_missed:
            raise ValueError("Input or target test samples are missed. They "
                             "must be defined together or none of them.")

        input_train = self.format_input_data(input_train)
        target_train = self.format_target_data(target_train)

        if input_test is not None:
            input_test = self.format_input_data(input_test)

        if target_test is not None:
            target_test = self.format_target_data(target_test)

        return super(ConstructibleNetwork, self).train(
            input_train=input_train, target_train=target_train,
            input_test=input_test, target_test=target_test,
            *args, **kwargs
        )

    def train_epoch(self, input_train, target_train):
        """
        Trains neural network over one epoch.

        Parameters
        ----------
        input_data : array-like
        target_data : array-like

        Returns
        -------
        float
            Prediction error.
        """
        return self.methods.train_epoch(*as_tuple(input_train, target_train))

    def architecture(self):
        """
        Shows network's architecture in the terminal if
        ``verbose`` parameter is equal to ``True``.
        """
        if not is_sequential(self.connection):
            raise TypeError("You can check architecture only for sequential "
                            "connections. For other types of connections it's "
                            "better to use the `neupy.plots.layer_structure` "
                            "function.")

        self.logs.title("Network's architecture")

        values = []
        for index, layer in enumerate(self.layers, start=1):
            input_shape = preformat_layer_shape(layer.input_shape)
            output_shape = preformat_layer_shape(layer.output_shape)
            classname = layer.__class__.__name__

            values.append((index, input_shape, classname, output_shape))

        self.logs.table(values, headers=[
            '#', 'Input shape', 'Layer type', 'Output shape'
        ])
        self.logs.newline()

    def __repr__(self):
        return "{}({}, {})".format(self.class_name(), self.connection,
                                   self.repr_options())



import theano
import theano.tensor as T
import numpy as np

from neupy.utils import asfloat
from neupy.algorithms.gd import StepSelectionBuiltIn
from neupy.core.properties import BoundedProperty, ProperFractionProperty
from .base import GradientDescent


__all__ = ('RPROP', 'IRPROPPlus')


class RPROP(StepSelectionBuiltIn, GradientDescent):
    """
    Resilient backpropagation (RPROP) is an optimization
    algorithm for supervised learning.

    Parameters
    ----------
    minstep : float
        Minimum possible value for step. Defaults to ``0.1``.

    maxstep : float
        Maximum possible value for step. Defaults to ``50``.

    increase_factor : float
        Increase factor for step in case when gradient doesn't change
        sign compare to previous epoch.

    decrease_factor : float
        Decrease factor for step in case when gradient changes sign
        compare to previous epoch.

    {GradientDescent.Parameters}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> rpropnet = algorithms.RPROP((2, 3, 1))
    >>> rpropnet.train(x_train, y_train)

    See Also
    --------
    :network:`IRPROPPlus` : iRPROP+ algorithm.
    :network:`GradientDescent` : GradientDescent algorithm.
    """

    # This properties correct upper and lower bounds for steps.
    minstep = BoundedProperty(default=0.1, minval=0)
    maxstep = BoundedProperty(default=50, minval=0)

    # This properties increase/decrease step by deviding it to
    # some coeffitient.
    increase_factor = BoundedProperty(minval=1, default=1.2)
    decrease_factor = ProperFractionProperty(default=0.5)

    def init_prev_delta(self, parameter):
        parameter_shape = T.shape(parameter).eval()
        self.prev_delta = theano.shared(
            name="{}/prev-delta".format(parameter.name),
            value=asfloat(np.zeros(parameter_shape)),
        )
        return self.prev_delta

    def init_param_updates(self, layer, parameter):
        prev_delta = self.init_prev_delta(parameter)

        parameter_shape = T.shape(parameter).eval()
        steps = theano.shared(
            name="{}/steps".format(parameter.name),
            value=asfloat(np.ones(parameter_shape) * self.step),
        )
        prev_gradient = theano.shared(
            name="{}/prev-grad".format(parameter.name),
            value=asfloat(np.zeros(parameter_shape)),
        )

        gradient = T.grad(self.variables.error_func, wrt=parameter)

        grad_product = prev_gradient * gradient
        negative_gradients = T.lt(grad_product, 0)

        updated_steps = T.clip(
            T.switch(
                T.gt(grad_product, 0),
                steps * self.increase_factor,
                T.switch(
                    negative_gradients,
                    steps * self.decrease_factor,
                    steps
                )
            ),
            self.minstep,
            self.maxstep,
        )
        gradient_signs = T.switch(T.lt(gradient, 0), -1, 1)
        parameter_delta = T.switch(
            negative_gradients,
            prev_delta,
            gradient_signs * updated_steps
        )
        updated_prev_gradient = T.switch(negative_gradients, 0, gradient)

        return [
            (parameter, parameter - parameter_delta),
            (steps, updated_steps),
            (prev_gradient, updated_prev_gradient),
            (self.prev_delta, -parameter_delta),
        ]


class IRPROPPlus(RPROP):
    """
    iRPROP+ is an optimization algorithm for supervised learning.
    This is a variation of the :network:`RPROP` algorithm.

    Parameters
    ----------
    {RPROP.minstep}

    {RPROP.maxstep}

    {RPROP.increase_factor}

    {RPROP.decrease_factor}

    {GradientDescent.addons}

    {ConstructibleNetwork.connection}

    {ConstructibleNetwork.error}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Methods
    -------
    {BaseSkeleton.predict}

    {ConstructibleNetwork.train}

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> rpropnet = algorithms.IRPROPPlus((2, 3, 1))
    >>> rpropnet.train(x_train, y_train)

    See Also
    --------
    :network:`RPROP` : RPROP algorithm.
    :network:`GradientDescent` : GradientDescent algorithm.
    """

    def init_variables(self):
        super(IRPROPPlus, self).init_variables()
        self.variables.update(
            last_error=theano.shared(name='irprop-plus/last-error',
                                     value=np.nan),
            previous_error=theano.shared(name='irprop-plus/previous-error',
                                         value=np.nan),
        )

    def on_epoch_start_update(self, epoch):
        super(IRPROPPlus, self).on_epoch_start_update(epoch)

        previous_error = self.errors.previous()
        if previous_error:
            last_error = self.errors.last()
            self.variables.last_error.set_value(last_error)
            self.variables.previous_error.set_value(previous_error)

    def init_prev_delta(self, parameter):
        prev_delta = super(IRPROPPlus, self).init_prev_delta(parameter)

        last_error = self.variables.last_error
        prev_error = self.variables.previous_error

        return T.switch(last_error > prev_error, prev_delta, 0)



import numpy as np

from neupy import init
from neupy.utils import format_data
from neupy.core.properties import IntProperty, ParameterProperty, ArrayProperty
from neupy.algorithms.base import BaseNetwork


__all__ = ('BaseStepAssociative',)


class BaseAssociative(BaseNetwork):
    """
    Base class for associative learning.

    Parameters
    ----------
    n_inputs : int
        Number of input units.

    n_outputs : int
        Number of output units.

    weight : array-like, Initializer
        Neural network weights.
        Value defined manualy should have shape ``(n_inputs, n_outputs)``.
        Defaults to :class:`Normal() <neupy.init.Normal>`.

    {BaseNetwork.step}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Methods
    -------
    {BaseSkeleton.predict}

    train(input_train, summary='table', epochs=100)
        Train neural network.

    {BaseSkeleton.fit}
    """
    n_inputs = IntProperty(minval=1, required=True)
    n_outputs = IntProperty(minval=1, required=True)
    weight = ParameterProperty(default=init.Normal())

    def __init__(self, **options):
        super(BaseAssociative, self).__init__(**options)
        self.init_layers()

    def init_layers(self):
        valid_weight_shape = (self.n_inputs, self.n_outputs)

        if isinstance(self.weight, init.Initializer):
            self.weight = self.weight.sample(valid_weight_shape)

        if self.weight.shape != valid_weight_shape:
            raise ValueError("Weight matrix has invalid shape. Got {}, "
                             "expected {}".format(self.weight.shape,
                                                  valid_weight_shape))

        self.weight = self.weight.astype(float)

    def format_input_data(self, input_data):
        is_feature1d = self.n_inputs == 1
        input_data = format_data(input_data, is_feature1d)

        if input_data.ndim != 2:
            raise ValueError("Cannot make prediction, because input "
                             "data has more than 2 dimensions")

        n_samples, n_features = input_data.shape

        if n_features != self.n_inputs:
            raise ValueError("Input data expected to have {} features, "
                             "but got {}".format(self.n_inputs, n_features))

        return input_data

    def train(self, input_train, summary='table', epochs=100):
        input_train = self.format_input_data(input_train)

        return super(BaseAssociative, self).train(
            input_train=input_train, target_train=None,
            input_test=None, target_test=None,
            epochs=epochs, epsilon=None,
            summary=summary)


class BaseStepAssociative(BaseAssociative):
    """
    Base class for associative algorithms which have 2 layers and first
    one is has step function as activation.

    Parameters
    ----------
    {BaseAssociative.n_inputs}

    {BaseAssociative.n_outputs}

    n_unconditioned : int
        Number of unconditioned units in neraul networks. All these
        units wouldn't update during the training procedure.
        Unconditioned should be the first feature in the dataset.

    weight : array-like
        Neural network weights.
        Value defined manualy should have shape ``(n_inputs, n_outputs)``.
        Defaults to ``None`` which means that all unconditional
        weights will be equal to ``1``. Other weights equal to ``0``.

    bias : array-like, Initializer
        Neural network bias units.
        Defaults to :class:`Constant(-0.5) <neupy.init.Constant>`.

    {BaseNetwork.step}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Methods
    -------
    {BaseSkeleton.predict}

    {BaseAssociative.train}

    {BaseSkeleton.fit}
    """
    n_inputs = IntProperty(minval=2, required=True)
    n_unconditioned = IntProperty(minval=1, required=True)

    weight = ArrayProperty()
    bias = ParameterProperty(default=init.Constant(-0.5))

    def init_layers(self):
        if self.n_inputs <= self.n_unconditioned:
            raise ValueError(
                "Number of uncondition features should be less than total "
                "number of features. `n_inputs`={} and "
                "`n_unconditioned`={}".format(
                    self.n_inputs,
                    self.n_unconditioned
                )
            )

        valid_weight_shape = (self.n_inputs, self.n_outputs)
        valid_bias_shape = (self.n_outputs,)

        if self.weight is None:
            self.weight = np.zeros(valid_weight_shape)
            self.weight[:self.n_unconditioned, :] = 1

        if isinstance(self.bias, init.Initializer):
            self.bias = self.bias.sample(valid_bias_shape)

        super(BaseStepAssociative, self).init_layers()

        if self.bias.shape != valid_bias_shape:
            raise ValueError("Bias vector has invalid shape. Got {}, "
                             "expected {}".format(self.bias.shape,
                                                  valid_bias_shape))

        self.bias = self.bias.astype(float)

    def predict(self, input_data):
        input_data = format_data(input_data, is_feature1d=False)
        raw_output = input_data.dot(self.weight) + self.bias
        return np.where(raw_output > 0, 1, 0)

    def train(self, input_train, *args, **kwargs):
        input_train = format_data(input_train, is_feature1d=False)
        return super(BaseStepAssociative, self).train(
            input_train, *args, **kwargs)

    def train_epoch(self, input_train, target_train):
        weight = self.weight
        n_unconditioned = self.n_unconditioned
        predict = self.predict
        weight_delta = self.weight_delta

        for input_row in input_train:
            input_row = np.reshape(input_row, (1, input_row.size))
            layer_output = predict(input_row)
            weight[n_unconditioned:, :] += weight_delta(input_row,
                                                        layer_output)



import json
import pkgutil
import importlib
from time import gmtime, strftime

import six
import theano
import numpy as np
from six.moves import cPickle as pickle

import neupy
from neupy.utils import asfloat
from neupy.core.docs import shared_docs
from neupy.layers.utils import extract_connection


__all__ = (
    'save', 'load',  # aliases to pickle
    'save_pickle', 'load_pickle',
    'save_json', 'load_json',
    'save_hdf5', 'load_hdf5',
    'load_dict', 'save_dict')


class ParameterLoaderError(Exception):
    """
    Exception triggers in case if there are some issues
    during the parameter loading procedure.
    """


class InvalidFormat(Exception):
    """
    Exception triggers when there are some issue with
    data format that stores connection data.
    """


def load_hdf5_module():
    """
    Loads `h5py` module. THis module used for manipulations
    with HDF5 files.

    Raises
    ------
    ImportError
        In case if module `h5py` wasn't installed

    Returns
    -------
    module
    """
    if not pkgutil.find_loader('h5py'):
        raise ImportError(
            "The `h5py` library wasn't installed. Try to "
            "install it with pip: \n    pip install h5py")

    return importlib.import_module('h5py')


def validate_layer_compatibility(layer, layer_data):
    """
    Checkes if it's possible to load stored data in the
    specified layer.

    Raises
    ------
    ParameterLoaderError
        When there are some problems with stored data.
    """
    expected_input_shape = layer_data['input_shape']
    expected_output_shape = layer_data['output_shape']

    if list(expected_input_shape) != list(layer.input_shape):
        raise ParameterLoaderError(
            "Layer `{}` expected to have input shape equal to {}, got {}"
            "".format(layer.name, expected_input_shape, layer.input_shape))

    if list(expected_output_shape) != list(layer.output_shape):
        raise ParameterLoaderError(
            "Layer `{}` expected to have output shape equal to {}, got {}"
            "".format(layer.name, expected_output_shape,
                      layer.output_shape))


def load_layer_parameter(layer, layer_data):
    """
    Set layer parameters to the values specified in the
    stored data
    """
    for param_name, param_data in layer_data['parameters'].items():
        parameter = getattr(layer, param_name)
        parameter.set_value(asfloat(param_data['value']))


def load_dict_by_names(layers_conn, layers_data, ignore_missed=False):
    """"
    Load parameters in to layer using layer names as the reference.

    Raises
    ------
    ParameterLoaderError
        In case if it's impossible to load parameters from data

    Returns
    -------
    bool
        Returns `True` in case if data was loaded successfully
        and `False` when parameters wasn't loaded
    """
    layers_data = {l['name']: l for l in layers_data}
    layers_conn = {l.name: l for l in layers_conn}

    if not ignore_missed and layers_data.keys() != layers_conn.keys():
        raise ParameterLoaderError(
            "Cannot match layers by name. \n"
            "  Layer names in connection: {}\n"
            "  Layer names in stored data: {}"
            "".format(layers_conn.keys(), layers_data.keys()))

    elif ignore_missed and all(l not in layers_data for l in layers_conn):
        raise ParameterLoaderError("Non of the layers can be matched by name")

    for layer_name, layer in layers_conn.items():
        if layer_name in layers_data:
            validate_layer_compatibility(layer, layers_data[layer_name])

    for layer_name, layer in layers_conn.items():
        if layer_name in layers_data:
            load_layer_parameter(layer, layers_data[layer_name])


def load_dict_sequentially(layers_conn, layers_data):
    """"
    Load parameters in to layer using sequential order of
    layer in connection and stored data
    """
    # It's important to point out that it can be that there more
    # stored layers than specified in the network. For this case we
    # expect to match as much as we can in case if layer are matchable.
    for layer, layer_data in zip(layers_conn, layers_data):
        validate_layer_compatibility(layer, layer_data)

    for layer, layer_data in zip(layers_conn, layers_data):
        load_layer_parameter(layer, layer_data)


def validate_data_structure(data):
    """
    Validates structure of the stored data

    Parameters
    ----------
    data : dict

    Raises
    ------
    InvalidFormat
        When format is invalid
    """
    if not isinstance(data, dict):
        raise InvalidFormat("Stored data should be in dictionary format")

    if 'layers' not in data:
        raise InvalidFormat("Stored data has no key `layers`")

    if not isinstance(data['layers'], list):
        raise InvalidFormat("Layers stored not in the list format")

    if not data['layers']:
        raise InvalidFormat("Stored data don't have any layer stored")

    for layer_index, layer in enumerate(data['layers']):
        if not isinstance(layer, dict):
            raise InvalidFormat(
                "Layer in the {} position (0-based indeces) is not a "
                "dictionary (it is {})".format(layer_index, type(layer)))

        for attr in ('parameters', 'input_shape', 'output_shape', 'name'):
            if attr not in layer:
                raise InvalidFormat(
                    "Layer in the {} position (0-based indeces) don't "
                    "have key `{}` specified".format(layer_index, attr))

        for attr in ('input_shape', 'output_shape'):
            if not isinstance(layer[attr], (list, tuple)):
                raise InvalidFormat(
                    "{} has invalid format for shape. It should be list "
                    "or tuple, got {}".format(attr, type(layer[attr])))

        if not isinstance(layer['parameters'], dict):
            raise InvalidFormat(
                "Layer in the {} position (0-based indeces) parameters "
                "specified as `{}`, but dictionary expected"
                "".format(layer_index, type(layer['parameters'])))

        for param_name, param in layer['parameters'].items():
            if not isinstance(param, dict):
                raise InvalidFormat(
                    "Layer in the {} position (0-based indeces) has "
                    "incorrect value for parameter named `{}`. It has "
                    "to be a dictionary, but got {}"
                    "".format(layer_index, param_name, type(param)))

            if 'value' not in param:
                raise InvalidFormat(
                    "Layer in the {} position (0-based indeces) has "
                    "incorrect value for parameter named `{}`. Parameter "
                    "doesn't have key named `value`"
                    "".format(layer_index, param_name))


def load_dict(connection, data, ignore_missed=False, load_by='names_or_order'):
    """
    Load network connections from dictionary.

    Parameters
    ----------
    connection : network, list of layer or connection

    data : dict
        Dictionary that stores network parameters.

    ignore_missed : bool
        ``False`` means that error will be triggered in case
        if some of the layers doesn't have storage parameters
        in the specified source. Defaults to ``False``.

    load_by : {{``names``, ``order``, ``names_or_order``}}
        Defines strategy that will be used during parameter loading

        - ``names`` - Matches layers in the network with stored layer
          using their names.

        - ``order`` - Matches layers in the network with stored layer
          using exect order of layers.

        - ``names_or_order`` - Matches layers in the network with stored
          layer trying to do it first using the same names and then
          matching them sequentialy.

        Defaults to ``names_or_order``.

    Raises
    ------
    ValueError
        Happens in case if `ignore_missed=False` and there is no
        parameters for some of the layers.
    """
    if load_by not in ('names', 'order', 'names_or_order'):
        raise ValueError(
            "Invalid value for the `load_by` argument: {}. Should be "
            "one of the following values: names, order, names_or_order."
            "".format(load_by))

    validate_data_structure(data)
    connection = extract_connection(connection)

    # We are only interested in layers that has parameters
    layers = data['layers']
    layers_data = [l for l in layers if l['parameters']]
    layers_conn = [l for l in connection if l.parameters]

    if not ignore_missed and len(layers_data) != len(layers_conn):
        raise ParameterLoaderError(
            "Couldn't load parameters from the dictionary. Connection "
            "has {} layers with parameters whether stored data has {}"
            "".format(len(layers_data), len(layers_conn)))

    if load_by == 'names':
        load_dict_by_names(layers_conn, layers_data, ignore_missed)

    elif load_by == 'order':
        load_dict_sequentially(layers_conn, layers_data)

    else:
        try:
            # First we try to load parameters using there names as
            # identifiers. Names are more reliable identifiers than
            # order of layers in the network
            load_dict_by_names(layers_conn, layers_data, ignore_missed)

        except ParameterLoaderError:
            # If we couldn't load data using layer names we will try to
            # compare layers in sequence one by one. Even if names are
            # different networks can be the same and order of parameters
            # should also be the same
            load_dict_sequentially(layers_conn, layers_data)

    # We need to initalize connection, to make sure
    # that each layer will generate shared variables
    # and validate connections
    connection.initialize()


def save_dict(connection):
    """
    Save network into the dictionary.

    Parameters
    ----------
    connection : network, list of layer or connection

    Returns
    -------
    dict
        Saved parameters and information about network in dictionary
        using specific format. Learn more about the NeuPy's storage
        format in the official documentation.

    Examples
    --------
    >>> from neupy import layers, storage
    >>>
    >>> connection = layers.Input(10) > layers.Softmax(3)
    >>> layers_data = storage.save_dict(connection)
    >>>
    >>> layers_data.keys()
    ['layers', 'graph', 'metadata']
    """
    connection = extract_connection(connection)
    data = {
        'metadata': {
            'language': 'python',
            'library': 'neupy',
            'version': neupy.__version__,
            'created': strftime("%a, %d %b %Y %H:%M:%S %Z", gmtime()),
            'theano_float': theano.config.floatX,
        },
        # Make it as a list in order to save the right order
        # of paramters, otherwise it can be convert to the dictionary.
        'graph': connection.graph.layer_names_only(),
        'layers': [],
    }

    for layer in connection:
        parameters = {}
        configs = {}

        for attrname, parameter in layer.parameters.items():
            parameters[attrname] = {
                'value': asfloat(parameter.get_value()),
                'trainable': parameter.trainable,
            }

        for option_name in layer.options:
            if option_name not in parameters:
                configs[option_name] = getattr(layer, option_name)

        data['layers'].append({
            'class_name': layer.__class__.__name__,
            'input_shape': layer.input_shape,
            'output_shape': layer.output_shape,
            'name': layer.name,
            'parameters': parameters,
            'configs': configs,
        })

    return data


@shared_docs(save_dict)
def save_pickle(connection, filepath, python_compatible=True):
    """
    Save layer parameters in pickle file.

    Parameters
    ----------
    {save_dict.connection}

    filepath : str
        Path to the pickle file that stores network parameters.

    python_compatible : bool
        If `True` pickled object would be compatible with
        Python 2 and 3 (pickle protocol equalt to `2`).
        If `False` then value would be pickled as highest
        protocol (`pickle.HIGHEST_PROTOCOL`).
        Defaults to `True`.

    Examples
    --------
    >>> from neupy import layers, storage
    >>>
    >>> connection = layers.Input(10) > layers.Softmax(3)
    >>> storage.save_pickle(connection, '/path/to/parameters.pickle')
    """
    connection = extract_connection(connection)
    data = save_dict(connection)

    with open(filepath, 'wb+') as f:
        # Protocol 2 is compatible for both python versions
        protocol = pickle.HIGHEST_PROTOCOL if python_compatible else 2
        pickle.dump(data, f, protocol)


@shared_docs(load_dict)
def load_pickle(connection, filepath, ignore_missed=False,
                load_by='names_or_order'):
    """
    Load and set parameters for layers from the
    specified filepath.

    Parameters
    ----------
    {load_dict.connection}

    filepath : str
        Path to pickle file that will store network parameters.

    {load_dict.ignore_missed}

    {load_dict.load_by}

    Raises
    ------
    {load_dict.Raises}

    Examples
    --------
    >>> from neupy import layers, storage
    >>>
    >>> connection = layers.Input(10) > layers.Softmax(3)
    >>> storage.load_pickle(connection, '/path/to/parameters.pickle')
    """
    connection = extract_connection(connection)

    with open(filepath, 'rb') as f:
        if six.PY3:
            # Specify encoding for python 3 in order to be able to
            # read files that has been created in python 2
            data = pickle.load(f, encoding='latin1')  # skip coverage
        else:
            data = pickle.load(f)  # skip coverage

    load_dict(connection, data, ignore_missed, load_by)


@shared_docs(save_dict)
def save_hdf5(connection, filepath):
    """
    Save network parameters in HDF5 format.

    Parameters
    ----------
    {save_dict.connection}

    filepath : str
        Path to the HDF5 file that stores network parameters.

    Examples
    --------
    >>> from neupy import layers, storage
    >>>
    >>> connection = layers.Input(10) > layers.Softmax(3)
    >>> storage.save_hdf5(connection, '/path/to/parameters.hdf5')
    """
    hdf5 = load_hdf5_module()
    connection = extract_connection(connection)
    data = save_dict(connection)

    with hdf5.File(filepath, mode='w') as f:
        layer_names = []

        for layer in data['layers']:
            layer_name = layer['name']
            layer_group = f.create_group(layer_name)

            for attrname, attrvalue in layer.items():
                if attrname != 'parameters':
                    layer_group.attrs[attrname] = json.dumps(
                        attrvalue, default=repr)

            for param_name, param in layer['parameters'].items():
                dataset = layer_group.create_dataset(
                    param_name, data=param['value'])

                dataset.attrs['trainable'] = param['trainable']

            layer_names.append(layer_name)

        f.attrs['metadata'] = json.dumps(data['metadata'])
        f.attrs['graph'] = json.dumps(data['graph'])
        f.attrs['layer_names'] = json.dumps(layer_names)


@shared_docs(load_dict)
def load_hdf5(connection, filepath, ignore_missed=False,
              load_by='names_or_order'):
    """
    Load network parameters from HDF5 file.

    Parameters
    ----------
    {load_dict.connection}

    filepath : str
        Path to HDF5 file that will store network parameters.

    {load_dict.ignore_missed}

    {load_dict.load_by}

    Raises
    ------
    {load_dict.Raises}

    Examples
    --------
    >>> from neupy import layers, storage
    >>>
    >>> connection = layers.Input(10) > layers.Softmax(3)
    >>> storage.load_hdf5(connection, '/path/to/parameters.hdf5')
    """
    hdf5 = load_hdf5_module()
    connection = extract_connection(connection)
    data = {}

    with hdf5.File(filepath, mode='r') as f:
        data['metadata'] = json.loads(f.attrs['metadata'])
        data['graph'] = json.loads(f.attrs['graph'])
        data['layers'] = []

        layer_names = json.loads(f.attrs['layer_names'])

        for layer_name in layer_names:
            layer_group = f[layer_name]
            layer = {'name': layer_name}

            for attrname, attrvalue in layer_group.attrs.items():
                try:
                    layer[attrname] = json.loads(attrvalue)
                except ValueError:
                    layer[attrname] = attrvalue

            layer['parameters'] = {}
            for param_name, parameter in layer_group.items():
                layer['parameters'][param_name] = {
                    'value': parameter.value,
                    'trainable': parameter.attrs['trainable'],
                }

            data['layers'].append(layer)

    load_dict(connection, data, ignore_missed, load_by)


def convert_numpy_array_to_list_recursively(data):
    for key, value in data.items():
        if isinstance(value, dict):
            convert_numpy_array_to_list_recursively(value)

        elif isinstance(value, np.ndarray):
            data[key] = value.tolist()

        elif isinstance(value, list):
            for entity in value:
                if isinstance(entity, dict):
                    convert_numpy_array_to_list_recursively(entity)


def dump_with_fastest_json_module(data, f, indent=None):
    """
    Data will be dumped using `ujson` module in case if it installed,
    otherwise Python's built-in `json` module will be used.

    Parameters
    ----------
    data : dict

    f : file object
        Opened file that available for writing

    indent : int or None
        JSON indentation.
    """
    # Without extra data processor we won't be able to dump
    # numpy array into json without raising an error.
    # `json` will have issues with numpy array encoding
    # `ujson` will have issue with numpy float or int encoding
    convert_numpy_array_to_list_recursively(data)

    if pkgutil.find_loader('ujson'):
        ujson = importlib.import_module('ujson')

        if indent is None:
            # Indentation functionality behaves differently compare
            # to the one that Pytnon has in built-in json module.
            #
            # Note: indent=0 will give different output
            # for ujson and json
            indent = 0

        return ujson.dump(data, f, indent=indent)
    return json.dump(data, f, indent=indent, default=repr)


def load_with_fastest_json_module(f):
    """
    Data will be loaded using `ujson` module in case if it installed,
    otherwise Python's built-in `json` module will be used.

    Parameters
    ----------
    f : file object
        Opened file that available for reading
    """
    if pkgutil.find_loader('ujson'):
        ujson = importlib.import_module('ujson')
        return ujson.load(f)
    return json.load(f)


@shared_docs(save_dict)
def save_json(connection, filepath, indent=None):
    """
    Save network parameters in JSON format.

    Parameters
    ----------
    {save_dict.connection}

    filepath : str
        Path to the JSON file that stores network parameters.

    indent : int or None
        Indentation that would be specified for the output JSON.
        Intentation equal to `2` or `4` makes it easy to read raw
        text files. The `None` value disables indentation which means
        that everything will be stored compactly. Defaults to `None`.

    Notes
    -----
    Install `ujson` library in order to speed up saving and
    loading procedure.

    Examples
    --------
    >>> from neupy import layers, storage
    >>>
    >>> connection = layers.Input(10) > layers.Softmax(3)
    >>> storage.save_json(connection, '/path/to/parameters.json')
    """
    connection = extract_connection(connection)
    data = save_dict(connection)

    with open(filepath, 'w') as f:
        dump_with_fastest_json_module(data, f)


@shared_docs(load_dict)
def load_json(connection, filepath, ignore_missed=False,
              load_by='names_or_order'):
    """
    Load network parameters from JSON file.

    Parameters
    ----------
    {load_dict.connection}

    filepath : str
        Path to JSON file that will store network parameters.

    {load_dict.ignore_missed}

    {load_dict.load_by}

    Raises
    ------
    {load_dict.Raises}

    Notes
    -----
    Install `ujson` library in order to speed up saving and
    loading procedure.

    Examples
    --------
    >>> from neupy import layers, storage
    >>>
    >>> connection = layers.Input(10) > layers.Softmax(3)
    >>> storage.load_json(connection, '/path/to/parameters.json')
    """
    connection = extract_connection(connection)
    data = save_dict(connection)

    with open(filepath, 'r') as f:
        data = load_with_fastest_json_module(f)

    load_dict(connection, data, ignore_missed, load_by)


# Convenient aliases
save = save_pickle
load = load_pickle



import abc
import math

import six
import numpy as np

from neupy.core.docs import SharedDocsABCMeta


__all__ = ('Initializer', 'Constant', 'Normal', 'Uniform', 'Orthogonal',
           'HeNormal', 'HeUniform', 'XavierNormal', 'XavierUniform')


class UninitializedException(Exception):
    """
    Exception for uninitialized parameters.
    """


def identify_fans(shape):
    """
    Identify fans from shape.

    Parameters
    ----------
    shape : tuple or list
        Matrix shape.

    Returns
    -------
    tuple
        Tuple that contains :math:`fan_{in}` and :math:`fan_{out}`.
    """
    fan_in = shape[0]
    output_feature_shape = shape[1:]

    if output_feature_shape:
        fan_out = np.prod(output_feature_shape).item(0)
    else:
        fan_out = 1

    return fan_in, fan_out


def classname(instance):
    """
    Returns instance class name.

    Parameters
    ----------
    instance : object

    Returns
    -------
    str
    """
    return instance.__class__.__name__


class Initializer(six.with_metaclass(SharedDocsABCMeta)):
    """
    Base class for parameter initialization.

    Methods
    -------
    sample(shape)
        Returns tensor with specified shape.
    """
    inherit_method_docs = True

    @abc.abstractmethod
    def sample(self, shape):
        """
        Returns tensor with specified shape.

        Parameters
        ----------
        shape : tuple
            Parameter shape.

        Returns
        -------
        array-like
        """
        raise NotImplementedError

    def get_value(self):
        """
        This method is the same as ``get_value`` for the Theano
        shared variables. The main point is to be able to
        generate understandable message when user try to get
        value from the uninitialized parameter.
        """
        raise UninitializedException("Cannot get parameter value. "
                                     "Parameter hasn't been initialized yet.")

    def __repr__(self):
        return '{}()'.format(classname(self))


class Constant(Initializer):
    """
    Initialize parameter that has constant values.

    Parameters
    ----------
    value : float, int
        All parameters in the tensor will be equal to
        this value. Defaults to ``0``.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, value=0):
        self.value = value

    def sample(self, shape):
        return np.ones(shape) * self.value

    def __repr__(self):
        return '{}({})'.format(classname(self), self.value)


class Normal(Initializer):
    """
    Initialize parameter sampling from the normal
    distribution.

    Parameters
    ----------
    mean : int, float
        Mean of the normal distribution.

    std : int, float
        Standard deviation of the normal distribution.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, mean=0, std=0.01):
        self.mean = mean
        self.std = std

    def sample(self, shape):
        return np.random.normal(loc=self.mean, scale=self.std, size=shape)

    def __repr__(self):
        return '{}(mean={}, std={})'.format(classname(self),
                                            self.mean, self.std)


class Uniform(Initializer):
    """
    Initialize parameter sampling from the uniformal
    distribution.

    Parameters
    ----------
    minval : int, float
        Minimum possible value.

    maxval : int, float
        Maximum possible value.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, minval=0, maxval=1):
        self.minval = minval
        self.maxval = maxval

    def sample(self, shape):
        minval, maxval = self.minval, self.maxval
        return np.random.random(shape) * (maxval - minval) + minval

    def __repr__(self):
        return '{}({}, {})'.format(classname(self),
                                   self.minval, self.maxval)


class Orthogonal(Initializer):
    """
    Initialize matrix with orthogonal basis.

    Parameters
    ----------
    scale : int, float
        Scales output matrix by a specified factor.
        Defaults to ``1``.

    Raises
    ------
    ValueError
        In case if tensor shape has more than 2
        dimensions.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, scale=1):
        self.scale = scale

    def sample(self, shape):
        ndim = len(shape)

        if ndim not in (1, 2):
            raise ValueError("Shape attribute must have 1 or 2 dimensions. "
                             "Found {} dimensions".format(ndim))

        rand_matrix = np.random.randn(*shape)

        if ndim == 1:
            return rand_matrix

        nrows, ncols = shape
        u, _, v = np.linalg.svd(rand_matrix, full_matrices=False)
        ortho_base = u if nrows > ncols else v

        return self.scale * ortho_base[:nrows, :ncols]

    def __repr__(self):
        return '{}(scale={})'.format(classname(self), self.scale)


class InitializerWithGain(Initializer):
    """
    Initialization class that has gain property

    Parameters
    ----------
    gain : float or {{``relu``}}
        Multiplies scaling factor by speified gain.
        The ``relu`` values set up gain equal to :math:`\\sqrt{{2}}`.
        Defaults to ``1``.
    """
    def __init__(self, gain=1.0):
        if gain == 'relu':
            gain = math.sqrt(2)

        self.gain = gain
        super(InitializerWithGain, self).__init__()


class HeNormal(InitializerWithGain):
    """
    Kaiming He parameter initialization method based on the
    normal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Kaiming He, Xiangyu Zhan, Shaoqing Ren, Jian Sun.
        Delving Deep into Rectifiers: Surpassing Human-Level
        Performance on ImageNet Classification, 2015.
    """
    def sample(self, shape):
        fan_in, _ = identify_fans(shape)
        variance = 2. / fan_in
        std = self.gain * np.sqrt(variance)
        return np.random.normal(loc=0, scale=std, size=shape)


class HeUniform(InitializerWithGain):
    """
    Kaiming He parameter initialization method based on the
    uniformal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Kaiming He, Xiangyu Zhan, Shaoqing Ren, Jian Sun.
        Delving Deep into Rectifiers: Surpassing Human-Level
        Performance on ImageNet Classification, 2015.
    """
    def sample(self, shape):
        fan_in, _ = identify_fans(shape)
        variance = 6. / fan_in
        abs_max_value = self.gain * np.sqrt(variance)

        uniform = Uniform(minval=-abs_max_value, maxval=abs_max_value)
        return uniform.sample(shape)


class XavierNormal(InitializerWithGain):
    """
    Xavier Glorot parameter initialization method based on
    normal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Xavier Glorot, Y Bengio. Understanding the difficulty
        of training deep feedforward neural networks, 2010.
    """
    def sample(self, shape):
        fan_in, fan_out = identify_fans(shape)
        variance = 2. / (fan_in + fan_out)
        std = self.gain * np.sqrt(variance)
        return np.random.normal(loc=0, scale=std, size=shape)


class XavierUniform(InitializerWithGain):
    """
    Xavier Glorot parameter initialization method based
    on uniform distribution.

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Xavier Glorot, Y Bengio. Understanding the difficulty
        of training deep feedforward neural networks, 2010.
    """
    def sample(self, shape):
        fan_in, fan_out = identify_fans(shape)
        variance = 6. / (fan_in + fan_out)
        abs_max_value = self.gain * np.sqrt(variance)

        uniform = Uniform(minval=-abs_max_value, maxval=abs_max_value)
        return uniform.sample(shape)



import numpy as np

from .base import BaseAssociative


__all__ = ('Kohonen',)


class Kohonen(BaseAssociative):
    """
    Kohonen Neural Network used for unsupervised learning.

    Parameters
    ----------
    {BaseAssociative.n_inputs}

    {BaseAssociative.n_outputs}

    {BaseAssociative.weight}

    {BaseNetwork.step}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Methods
    -------
    {BaseSkeleton.predict}

    {BaseAssociative.train}

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms, layers, environment
    >>>
    >>> environment.reproducible()
    >>>
    >>> input_data = np.array([
    ...     [0.1961,  0.9806],
    ...     [-0.1961,  0.9806],
    ...     [0.9806,  0.1961],
    ...     [0.9806, -0.1961],
    ...     [-0.5812, -0.8137],
    ...     [-0.8137, -0.5812],
    ... ])
    >>>
    >>> kohonet = algorithms.Kohonen(
    ...     n_inputs=2,
    ...     n_outputs=3,
    ...     step=0.5,
    ...     verbose=False
    ... )
    >>> kohonet.train(input_data, epochs=100)
    >>> kohonet.predict(input_data)
    array([[ 0.,  1.,  0.],
           [ 0.,  1.,  0.],
           [ 1.,  0.,  0.],
           [ 1.,  0.,  0.],
           [ 0.,  0.,  1.],
           [ 0.,  0.,  1.]])
    """
    def predict_raw(self, input_data):
        input_data = self.format_input_data(input_data)
        return input_data.dot(self.weight)

    def predict(self, input_data):
        raw_output = self.predict_raw(input_data)
        output = np.zeros(raw_output.shape, dtype=np.int0)
        max_args = raw_output.argmax(axis=1)
        output[range(raw_output.shape[0]), max_args] = 1
        return output

    def train_epoch(self, input_train, target_train):
        step = self.step
        predict = self.predict

        error = 0
        for input_row in input_train:
            input_row = np.reshape(input_row, (1, input_row.size))
            layer_output = predict(input_row)

            _, index_y = np.nonzero(layer_output)
            distance = input_row.T - self.weight[:, index_y]
            self.weight[:, index_y] += step * distance

            error += np.abs(distance).mean()

        return error / len(input_train)



import theano.tensor as T


__all__ = ('preformat_layer_shape', 'dimshuffle', 'iter_parameters',
           'count_parameters', 'create_input_variable', 'extract_connection')


def preformat_layer_shape(shape):
    """
    Format layer's input or output shape.

    Parameters
    ----------
    shape : int or tuple

    Returns
    -------
    int or tuple
    """
    if isinstance(shape, tuple) and len(shape) == 1:
        return shape[0]
    return shape


def dimshuffle(value, ndim, axes):
    """
    Shuffle dimension based on the specified number of
    dimensions and axes.

    Parameters
    ----------
    value : Theano variable
    ndim : int
    axes : tuple, list

    Returns
    -------
    Theano variable
    """
    pattern = ['x'] * ndim

    for i, axis in enumerate(axes):
        pattern[axis] = i

    return value.dimshuffle(pattern)


def iter_parameters(layers, only_trainable=True):
    """
    Iterate through layer parameters.

    Parameters
    ----------
    layers : list of layers or connection

    only_trainable : bool
        If `True` returns only trainable parameters.
        Defaults to `True`.

    Yields
    ------
    tuple
        Tuple with three ariables: (layer, attribute_name, parameter)
    """
    observed_parameters = []

    for layer in layers:
        for attrname, parameter in layer.parameters.items():
            new_parameter = parameter not in observed_parameters

            if new_parameter and (parameter.trainable or not only_trainable):
                observed_parameters.append(parameter)
                yield layer, attrname, parameter


def count_parameters(connection):
    """
    Count number of parameters in Neural Network.

    Parameters
    ----------
    connection : list of laters or connection

    Returns
    -------
    int
        Number of parameters.
    """
    n_parameters = 0

    for _, _, parameter in iter_parameters(connection):
        parameter = parameter.get_value()
        n_parameters += parameter.size

    return n_parameters


def create_input_variable(input_shape, name):
    """
    Create input variable based on the specified
    input shape.

    Parameters
    ----------
    input_shape : tuple
    name : str

    Returns
    -------
    Theano variable
    """
    dim_to_variable_type = {
        2: T.matrix,
        3: T.tensor3,
        4: T.tensor4,
    }

    # Shape doesn't include batch size dimension,
    # that's why we need to add one
    ndim = len(input_shape) + 1

    if ndim not in dim_to_variable_type:
        raise ValueError("Layer's input needs to be 2, 3 or 4 "
                         "dimensional. Found {} dimensions".format(ndim))

    variable_type = dim_to_variable_type[ndim]
    return variable_type(name)


def extract_connection(instance):
    """
    Extract connection from different types of object.

    Parameters
    ----------
    instance : network, connection, list or tuple

    Returns
    -------
    connection

    Raises
    ------
    ValueError
        In case if input object doesn't have connection of layers.
    """
    # Note: Import it here in order to prevent loops
    from neupy.algorithms.base import BaseNetwork
    from neupy.layers.base import BaseConnection
    from neupy import layers

    if isinstance(instance, (list, tuple)):
        return layers.join(*instance)

    if isinstance(instance, BaseNetwork):
        return instance.connection

    if isinstance(instance, BaseConnection):
        return instance

    raise TypeError("Invalid input type. Input should be network, connection "
                    "or list of layers, got {}".format(type(instance)))



import numpy as np

from neupy.utils import format_data
from neupy.exceptions import NotTrained
from neupy.core.properties import IntProperty, ParameterProperty
from neupy.algorithms.base import BaseNetwork
from neupy import init


__all__ = ('Oja',)


class Oja(BaseNetwork):
    """
    Oja is an unsupervised technique used for the
    dimensionality reduction tasks.

    Notes
    -----
    - In practice use step as very small value.
      For instance, value ``1e-7`` can be a good choice.

    - Normalize the input data before use Oja algorithm.
      Input data shouldn't contains large values.

    - Set up smaller values for weight if error for a few
      first iterations is big compare to the input values scale.
      For instance, if your input data have values between
      ``0`` and ``1`` error value equal to ``100`` is big.

    Parameters
    ----------
    minimized_data_size : int
        Expected number of features after minimization,
        defaults to ``1``.

    weight : array-like or ``None``
        Defines networks weights.
        Defaults to :class:`XavierNormal() <neupy.init.XavierNormal>`.

    {BaseNetwork.step}

    {BaseNetwork.show_epoch}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Methods
    -------
    reconstruct(input_data)
        Reconstruct original dataset from the minimized input.

    train(input_data, epsilon=1e-2, epochs=100)
        Trains algorithm based on the input dataset.
        For the dimensionality reduction input dataset
        assumes to be also a target.

    {BaseSkeleton.predict}

    {BaseSkeleton.fit}

    Raises
    ------
    ValueError
        - Triggers when you try to reconstruct output
          without training.

        - Invalid number of input data features for the
          ``train`` and ``reconstruct`` methods.

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> data = np.array([[2, 2], [1, 1], [4, 4], [5, 5]])
    >>>
    >>> ojanet = algorithms.Oja(
    ...     minimized_data_size=1,
    ...     step=0.01,
    ...     verbose=False
    ... )
    >>>
    >>> ojanet.train(data, epsilon=1e-5)
    >>> minimized = ojanet.predict(data)
    >>> minimized
    array([[-2.82843122],
           [-1.41421561],
           [-5.65686243],
           [-7.07107804]])
    >>> ojanet.reconstruct(minimized)
    array([[ 2.00000046,  2.00000046],
           [ 1.00000023,  1.00000023],
           [ 4.00000093,  4.00000093],
           [ 5.00000116,  5.00000116]])
    """
    minimized_data_size = IntProperty(minval=1)
    weight = ParameterProperty(default=init.XavierNormal())

    def train_epoch(self, input_data, target_train):
        weight = self.weight

        minimized = np.dot(input_data, weight)
        reconstruct = np.dot(minimized, weight.T)
        error = input_data - reconstruct

        weight += self.step * np.dot(error.T, minimized)

        mae = np.sum(np.abs(error)) / input_data.size

        # Clean objects from the memory
        del minimized
        del reconstruct
        del error

        return mae

    def train(self, input_data, epsilon=1e-2, epochs=100):
        input_data = format_data(input_data)
        n_input_features = input_data.shape[1]

        if isinstance(self.weight, init.Initializer):
            weight_shape = (n_input_features, self.minimized_data_size)
            self.weight = self.weight.sample(weight_shape)

        if n_input_features != self.weight.shape[0]:
            raise ValueError(
                "Invalid number of features. Expected {}, got {}".format(
                    self.weight.shape[0],
                    n_input_features
                )
            )

        super(Oja, self).train(input_data, epsilon=epsilon, epochs=epochs)

    def reconstruct(self, input_data):
        if not isinstance(self.weight, np.ndarray):
            raise NotTrained("Network hasn't been trained yet")

        input_data = format_data(input_data)
        if input_data.shape[1] != self.minimized_data_size:
            raise ValueError(
                "Invalid input data feature space, expected "
                "{}, got {}.".format(
                    input_data.shape[1],
                    self.minimized_data_size
                )
            )

        return np.dot(input_data, self.weight.T)

    def predict(self, input_data):
        if not isinstance(self.weight, np.ndarray):
            raise NotTrained("Network hasn't been trained yet")

        input_data = format_data(input_data)
        return np.dot(input_data, self.weight)



from .gd.base import *
from .gd.lev_marq import *
from .gd.quasi_newton import *
from .gd.conjgrad import *
from .gd.hessian import *
from .gd.hessdiag import *
from .gd.rprop import *
from .gd.quickprop import *
from .gd.momentum import *
from .gd.adadelta import *
from .gd.adagrad import *
from .gd.rmsprop import *
from .gd.adam import *
from .gd.adamax import *

from .ensemble.dan import *

from .regularization.weight_decay import *
from .regularization.weight_elimination import *
from .regularization.max_norm import *

from .step_update.step_decay import *
from .step_update.search_then_converge import *
from .step_update.errdiff import *
from .step_update.leak_step import *
from .step_update.linear_search import *

from .memory.discrete_hopfield_network import *
from .memory.bam import *
from .memory.cmac import *

from .associative.oja import *
from .associative.hebb import *
from .associative.instar import *
from .associative.kohonen import *

from .competitive.sofm import *
from .competitive.art import *
from .competitive.lvq import *
from .competitive.growing_neural_gas import *

from .rbfn.pnn import *
from .rbfn.rbf_kmeans import *
from .rbfn.grnn import *

from .linear.lms import *
from .linear.modify_relaxation import *
from .linear.perceptron import *

from .rbm import *
